#!/usr/bin/perl
#
# This script computes the left and right stop priors on a given corpus
# in CoNLL 2006 format. More details can be found in the following paper:
#
# David Marecek, Milan Straka: Stop-probability estimates computed on a
# large corpus improve Unsupervised Dependency Parsing. In Proceedings
# of ACL 2013.
#
# Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
# Copyright 2013 by Milan Straka <fox@ucw.cz>.
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this script.  If not, see <http://www.gnu.org/licenses/>.

use strict;
use warnings;
use Getopt::Long;

use open qw/:std :utf8/;

my $TAG_COLUMN = 4;
my $FORM_COLUMN = 1;
my $MAX_WORDS = 30;
my $MIN_WORDS = 10;
my $MAX_NGRAM = 3;
my $TAGMAP = '';

GetOptions(
    "tag-column=i"    => \$TAG_COLUMN,
    "form-column=i"   => \$FORM_COLUMN,
    "max-words=i"     => \$MAX_WORDS,
    "min-words=i"     => \$MIN_WORDS,
    "max-ngram=i"     => \$MAX_NGRAM,
    "tagmap=s"        => \$TAGMAP,
);

my %count;
my %sent2tags;
my %examples;

my $words = "";
my $tags = "";
my $length = 0;

my %tag_cnt;
my $tag_total;
my %tag_form_cnt;

my $sent_counter = 0;

my %tag_map = ();
if ($TAGMAP) {
  open (my $f, '<', $TAGMAP) or die "Cannot open file $TAGMAP with the tagmap.";
  while (<$f>) {
    chomp;
    my ($form, $tag) = split /\t/;
    die "Form $form present multiple times in $TAGMAP" if exists $tag_map{$form};
    $tag_map{$form} = $tag;
  }
  close $f;
}

while (<>) {
    chomp;
    my $line = $_;
    if ($line =~ /^\d/) {
        my @items = split /\t/, $line;
        my ($form, $tag) = ($items[$FORM_COLUMN], $items[$TAG_COLUMN]);
        $tag = $tag_map{$form} if exists $tag_map{$form};
        $tag_cnt{$tag}++;
        $tag_total++;
        $tag_form_cnt{$tag}{lc($form)}++;
        $words .= lc($form)." ";
        $tags  .= $tag." ";
        $examples{$tag} = "" if !defined $examples{$tag};
        $examples{$tag} .= "$form " if length $examples{$tag} < 100;
        $length++;
    }
    else {
        $sent_counter++;
        print STDERR "$sent_counter " if $sent_counter % 10000 == 0;
        $words =~ s/\s$//;
        $tags =~ s/\s$//;
        if ($length <= $MAX_WORDS) {
            $count{$words}++;
            $sent2tags{$words} = $tags;
        }
        $words = "";
        $tags = "";
        $length = 0;
    }
}
print STDERR "The corpus has been read.\n";

my @stop;
my @total;
my $stop_all;
my $total_all;

$sent_counter = 0;

foreach my $sent (keys %count) {
    $sent_counter++;
    print STDERR "$sent_counter " if $sent_counter % 10000 == 0;
    my @words = split /\s/, $sent;
    my @tags = split /\s/, $sent2tags{$sent};
    next if @words < $MIN_WORDS;
    foreach my $n (0 .. $MAX_NGRAM - 1) {
        foreach my $i (0 .. $#words - $n) {
            my $ngram = join (" ", @tags[$i .. ($i+$n)]);
            $total[0]{$tags[$i]}++;
            $total[1]{$tags[$i+$n]}++;
            $total_all++;    
            my $new_sent = join (" ", map {$words[$_]} ((0 .. ($i - 1)), (($i + $n + 1) .. $#words)));
            if ($count{$new_sent}) {
                $stop[0]{$tags[$i]}++;
                $stop[1]{$tags[$i+$n]}++;
                $stop_all++;
            }
        }
    }
}

print STDERR "Computing smoothing and normalization constants...\n";
my $smooth = $stop_all / $total_all;
my $norm;
my $should_be = 4 / 3 * $tag_total;
my $max = 1;
my $min = 1;
my $x = 1;
while ($x == 1 || $max/$min > 1.1){
    my $sum = 0;
    foreach my $dir (0, 1) {
        foreach my $t (keys %tag_cnt) {
            my $s = (($stop[$dir]{$t} || 0) + $smooth) / ($total[$dir]{$t} + 1);
            $sum += $s * $tag_cnt{$t} / ($x + $s);
        }
    }
    if ($sum > $should_be) {
        if ($x == $max) {
            $min = $x; $x *= 2; $max = $x;
        }
        else {
            $min = $x; $x = ($max + $x) / 2;
        }
    }
    else {
        if ($x == $min) {
            $max = $x; $x /= 2; $min = $x;
        }
        else {
            $max = $x; $x = ($min + $x) / 2;
        }
    }
}
$norm = ($max + $min) / 2;
foreach my $dir (0, 1) {
    foreach my $t (keys %tag_cnt) {
        my $s = (($stop[$dir]{$t} || 0) + $smooth) / ($total[$dir]{$t} + 1);
        $stop[$dir]{$t} = $s / ($norm + $s);
    }
}
print STDERR "Smoothing: $smooth, normalization: $norm, reducible n-grams: $stop_all\n";


my @l_ordered = sort {$stop[0]{$b} <=> $stop[0]{$a}} keys %{$stop[0]};
my @r_ordered = sort {$stop[1]{$b} <=> $stop[1]{$a}} keys %{$stop[1]};

foreach my $t (@l_ordered) {
    print "L\t$t\t$stop[0]{$t}\n";
}
foreach my $t (@r_ordered) {
    print "R\t$t\t$stop[1]{$t}\n";
}

