// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include <fstream>

#include "common.h"
#include "corpus.h"
#include "utils/utils.h"

void sentence::add_node(int form, int lcform, int tag, int oritag) {
  int ord = nodes.size();
  if (ord >= (2<<15)) runtime_errorf("A sentence can have at most 2^15 words");
  if (max(tag, oritag) >= (2<<15)) runtime_errorf("There can be at most 2^15 tags");
  nodes.emplace_back(ord, form, lcform, tag, oritag);

  if (!ord)
    root = 0;
  else {
    assert(nodes[ord-1].parent < 0);
    nodes[ord].sibling[LEFT] = ord-1;
    nodes[ord-1].sibling[RGHT] = ord;
  }
}

void sentence::flatten() {
  if (nodes.empty()) return;

  root = 0;
  foreach(auto& n, nodes) {
    n.parent = -1;
    n.sibling[LEFT] = n.ord - 1;
    n.sibling[RGHT] = n.ord + 1 == (int)nodes.size() ? -1 : n.ord + 1;
    n.children[LEFT] = n.children[RGHT] = 0;
    n.outer_child[LEFT] = n.outer_child[RGHT] = -1;
    n.range[LEFT] = n.range[RGHT] = n.ord;
  }
}

void sentence::change_parent(node& n, int parent) {
  assert(parent != n.ord);

  if (n.parent == parent) return;

  // Update children, outer_child and range of current parent
  if (n.parent >= 0) {
    nodes[n.parent].children[n.ord > n.parent]--;
    if (n.sibling[LEFT] < 0) nodes[n.parent].outer_child[LEFT] = n.sibling[RGHT];
    if (n.sibling[RGHT] < 0) nodes[n.parent].outer_child[RGHT] = n.sibling[LEFT];
    for (int p = n.parent; p >= 0 && nodes[p].range[LEFT] == n.range[LEFT]; p = nodes[p].parent) nodes[p].range[LEFT] = n.range[RGHT] + 1;
    for (int p = n.parent; p >= 0 && nodes[p].range[RGHT] == n.range[RGHT]; p = nodes[p].parent) nodes[p].range[RGHT] = n.range[LEFT] - 1;
  } else if (root == n.ord) root = n.sibling[RGHT];
  // Delete myself from siblings
  if (n.sibling[LEFT] >= 0) nodes[n.sibling[LEFT]].sibling[RGHT] = n.sibling[RGHT];
  if (n.sibling[RGHT] >= 0) nodes[n.sibling[RGHT]].sibling[LEFT] = n.sibling[LEFT];

  // Set correct n.siblings
  if (n.parent >= 0 && parent == nodes[n.parent].parent) {
    assert((n.ord < n.parent && (nodes[n.parent].sibling[LEFT] < 0 || n.ord > nodes[n.parent].sibling[LEFT])) ||
           (n.ord > n.parent && (nodes[n.parent].sibling[RGHT] < 0 || n.ord < nodes[n.parent].sibling[RGHT])));
    if (n.ord < n.parent) n.sibling[RGHT] = n.parent, n.sibling[LEFT] = nodes[n.parent].sibling[LEFT];
    else n.sibling[LEFT] = n.parent, n.sibling[RGHT] = nodes[n.parent].sibling[RGHT];
  } else {
    assert(parent >= 0);
    assert((nodes[parent].outer_child[LEFT] < 0 && nodes[parent].outer_child[RGHT] < 0) ||
           nodes[parent].outer_child[LEFT] > n.ord ||
           nodes[parent].outer_child[RGHT] < n.ord);
    if (nodes[parent].outer_child[RGHT] < n.ord) n.sibling[LEFT] = nodes[parent].outer_child[RGHT], n.sibling[RGHT] = -1;
    else n.sibling[RGHT] = nodes[parent].outer_child[LEFT], n.sibling[LEFT] = -1;
  }
  // Add myself as a sibling of my siblings
  if (n.sibling[LEFT] >= 0) nodes[n.sibling[LEFT]].sibling[RGHT] = n.ord;
  if (n.sibling[RGHT] >= 0) nodes[n.sibling[RGHT]].sibling[LEFT] = n.ord;

  // Update children, outer_child and range of new parent
  if (parent >= 0) {
    nodes[parent].children[n.ord > parent]++;
    if (nodes[parent].outer_child[LEFT] < 0 || nodes[parent].outer_child[LEFT] > n.ord) nodes[parent].outer_child[LEFT] = n.ord;
    if (nodes[parent].outer_child[RGHT] < n.ord) nodes[parent].outer_child[RGHT] = n.ord;
    for (int p = parent; p >= 0 && nodes[p].range[LEFT] > n.range[LEFT]; p = nodes[p].parent) nodes[p].range[LEFT] = n.range[LEFT];
    for (int p = parent; p >= 0 && nodes[p].range[RGHT] < n.range[RGHT]; p = nodes[p].parent) nodes[p].range[RGHT] = n.range[RGHT];
  } else if (root < 0 || root > n.ord) root = n.ord;

  // Finally set the parent
  n.parent = parent;
}

void sentence::assert_valid() {
  if (!nodes.size()) assert(root < 0);

  // There is at least one root and they form list via siblings.
  assert(root >= 0);
  int last_root = -1;
  int range[DIRS] = {-1, -1};
  foreach (auto& n, nodes)
    if (n.parent < 0) {
      if (last_root < 0) assert(root == n.ord);
      assert(n.sibling[LEFT] == last_root);
      if (last_root >= 0) assert(nodes[last_root].sibling[RGHT] == n.ord);

      // Recursive check
      assert_valid(n);
      if (range[RGHT] >= 0) assert(range[RGHT] == n.range[LEFT] - 1);
      else range[LEFT] = n.range[LEFT];
      range[RGHT] = n.range[RGHT];

      last_root = n.ord;
    }
  assert(last_root >= 0 && nodes[last_root].sibling[RGHT] < 0);
  assert(range[LEFT] == 0 && range[RGHT] == (int)nodes.size()-1);

  // Currently allow only one root
  assert(nodes[root].sibling[LEFT] < 0 && nodes[root].sibling[RGHT] < 0);
}

void sentence::assert_valid(const node &n) {
  int range[DIRS] = {-1, -1}, children[DIRS] = {0, 0};

  int last_child = -1;
  assert(n.outer_child[LEFT] < 0 || nodes[n.outer_child[LEFT]].parent == n.ord);
  assert(n.outer_child[RGHT] < 0 || nodes[n.outer_child[RGHT]].parent == n.ord);
  foreach (auto& c, nodes)
    if (c.parent == n.ord) {
      assert(last_child >= 0 || n.outer_child[LEFT] == c.ord);
      assert(c.sibling[LEFT] == last_child);
      assert(last_child < 0 || nodes[last_child].sibling[RGHT] == c.ord);

      assert_valid(c);

      if (range[RGHT] >= 0) assert(range[RGHT] == c.range[LEFT] - 1);
      else range[LEFT] = c.range[LEFT];
      range[RGHT] = c.range[RGHT];

      children[c.ord > n.ord]++;
      last_child = c.ord;
    } else if (c.ord == n.ord) {
      if (range[RGHT] >= 0) assert(range[RGHT] == c.ord - 1);
      else range[LEFT] = c.ord;
      range[RGHT] = c.ord;
    }
  if (last_child < 0) assert(n.outer_child[LEFT] < 0);
  assert(last_child == n.outer_child[RGHT]);
  if (last_child >= 0) assert(nodes[last_child].sibling[RGHT] < 0);

  assert(range[LEFT] == n.range[LEFT] && range[RGHT] == n.range[RGHT]);
  assert(children[LEFT] == n.children[LEFT] && children[RGHT] == n.children[RGHT]);
}

int corpus::form_id(const str& form) {
  auto it = forms_map.find(form);
  if (it == forms_map.end()) {
    forms.push_back(form);
    it = forms_map.insert({form, forms_map.size()}).first;
  }
  return it->second;
}

int corpus::lcform_id(const str& form) {
  str lcform;
  lowercase(form, lcform);
  auto it = lcforms_map.find(lcform);
  if (it == lcforms_map.end()) {
    lcforms.push_back(lcform);
    it = lcforms_map.insert({lcform, lcforms_map.size()}).first;
  }
  return it->second;
}

int corpus::tag_id(const str& tag) {
  auto it = tags_map.find(tag);
  if (it == tags_map.end()) {
    tags.push_back(tag);
    it = tags_map.insert({tag, tags_map.size()}).first;
  }
  return it->second;
}

void corpus::load_tagmap(istream& in) {
  str line;
  vector<str> tokens;

  while (gettextline(in, line)) {
    split(line, '\t', tokens);
    if (tokens.size() < 2) runtime_errorf("The tagmap file line '%s' does not contain at least 2 columns", line.c_str());
    tagmap.insert({tokens[0], tokens[1]});
  }
}

void corpus::load(istream& in, int form_column, int tag_column) {
  bool new_sentence = true;
  str line;
  vector<str> tokens;

  while (gettextline(in, line)) {
    if (line.empty())
      new_sentence = true;
    else {
      if (new_sentence) {
        sentences.push_back(sentence(sentences.size()));
        new_sentence = false;
      }

      split(line, '\t', tokens);
      if (max(form_column, tag_column) >= (int)tokens.size()) runtime_errorf("The input corpus line '%s' does not contain enough columns", line.c_str());
      sentences.back().add_node(form_id(tokens[form_column]), lcform_id(tokens[form_column]),
                                tag_id(tagmap.count(tokens[form_column]) ? tagmap[tokens[form_column]] : tokens[tag_column]),
                                tag_id(tokens[tag_column]));
    }
  }
}

void corpus::save(ostream& out) {
  foreach (auto& s, sentences) {
    foreach (auto& n, s.nodes)
      out << n.ord + 1 << "\t" << forms[n.form] << "\t_\t_\t" << tags[n.tag] << "\t_\t" << n.parent + 1 << "\t_\t_\t_\n";
    out << endl;
  }
}

void corpus::assert_valid() {
  foreach (auto& s, sentences)
    s.assert_valid();
}
