// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include "common.h"
#include "gibbs_sampler.h"
#include "graph/directed_mst.h"
#include "utils/distribution.h"

void gibbs_sampler::add_model(model *m) {
  models.push_back(m);
}

gibbs_sampler::~gibbs_sampler() {
  foreach (auto& m, models) delete m;
}

void gibbs_sampler::initialize_sentence(sentence &s, int left, int right) {
  int root = distribution::uniform(left, right);

  for (int ni = root-1; ni >= left; ni--) s.change_parent(s.nodes[ni], root);
  for (int ni = root+1; ni < right; ni++) s.change_parent(s.nodes[ni], root);

  if (root - left > initialize_max_children) initialize_sentence(s, left, root);
  if (right - root > initialize_max_children) initialize_sentence(s, root + 1, right);
}

double gibbs_sampler::initialize_models(bool compute_logprob) {
  double result = 0;

  foreach (auto& m, models)
    m->initialize(c, compute_logprob ? &result : NULL);

  return result;
}

void gibbs_sampler::initialize() {
  // Initialize corpus
  foreach (auto& s, c.sentences) {
    s.flatten();
    initialize_sentence(s, 0, s.nodes.size());
  }

  // Init counts
  sampled.clear();
  foreach (auto& s, c.sentences)
    sampled.emplace_back(s.nodes.size());

  // Construct models
  foreach (auto& m, models)
    m->construct(c);

  // Initialize samplers
  incremental_sampler.construct(models, verbose);
  block_sampler.construct(models, verbose);

  // Initialize models
  if (verbose) cerr << "Running Gibbs sampling, initial logprob is: ", cerr << initialize_models() << endl;
  else initialize_models(false);
}

void gibbs_sampler::sample(sampling_mode mode, int iterations, int num_samples) {
  int progress_step = mode == INCREMENTAL ? incremental_progress_step : block_progress_step;

  int sentences = c.sentences.size();
  vector<int> s_perm;
  for (int iter = 0; iter < iterations; iter++) {
    if (verbose) cerr << "Iteration " << iter+1 << "... ";
    distribution::permutation(0, sentences, s_perm);
    for (int si = 0; si < sentences; si++) {
      auto& s = c.sentences[s_perm[si]];

      switch(mode) {
        case INCREMENTAL: incremental_sampler.sample(s, iter); break;
        case BLOCK_MAX: block_sampler.sample(s, true, true); break;
        case BLOCK_RND: block_sampler.sample(s, false, true); break;
        case BLOCK_MAX_NOUPDATE: block_sampler.sample(s, true, false); break;
        case BLOCK_RND_NOUPDATE: block_sampler.sample(s, false, false); break;
      }

      if (verbose && si && (si & (progress_step-1)) == 0)
        cerr << " ["[si == progress_step] << 100 * si / sentences << '%';
    }

    if (verbose && sentences > progress_step) cerr << " 100%] ";
    if (verbose) cerr << "done, logprob: " << initialize_models() << endl;
    else initialize_models(false);

    if (iter >= iterations - num_samples) {
      // Samples for the MST decoder
      foreach (auto& s, c.sentences)
        foreach (auto& n, s.nodes)
          sampled[s.ord].counts[1 + n.ord][1 + n.parent]--;
      // Samples for BEST_PROB decoder
      foreach (auto& m, models)
        m->sample_state();
    }

    assert((c.assert_valid(), true));
  }
}

void gibbs_sampler::decode(decoding_mode mode) {
  if (verbose) cerr << "Decoding... ";
  if (mode == MST) {
    vector<int> parents;
    foreach (auto& s, c.sentences) {
      directed_mst(sampled[s.ord].counts, parents);

      s.flatten();
      foreach (auto& n, s.nodes)
        s.change_parent(n, parents[1 + n.ord] - 1);
    }
  } else /* if (mode == BEST_PROB) */ {
    foreach (auto& m, models)
      m->use_sampled_state();
    foreach (auto& s, c.sentences)
      block_sampler.sample(s, true, false);
  }
  if (verbose) cerr << "done." << endl;
  cerr << "Final logprob: " << initialize_models() << endl;
}
