// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#pragma GCC diagnostic ignored "-Warray-bounds"

#include "common.h"
#include "gibbs_sampler_block.h"
#include "utils/distribution.h"

void gibbs_sampler_block::construct(const vector<model*>& models, bool /*verbose*/) {
  foreach (auto& m, models) {
    auto model = m->block();
    if (!model) continue;

    block_models.push_back(model);
    if (model->use_edge_prob()) edge_models.push_back(model);
    if (model->use_range_prob()) range_models.push_back(model);
    if (model->use_stop_prob()) stop_models.push_back(model);
  }
}

void gibbs_sampler_block::sample(sentence& s, bool best, bool update_counts) {
  if (update_counts) foreach (auto& m, block_models) m->count_sentence(s, -1);
  switch (best) {
    case true: sample<Combine_max, Choose_max>(s); break;
    case false: sample<Combine_rnd, Choose_rnd>(s); break;
  }
  if (update_counts) foreach (auto& m, block_models) m->count_sentence(s, +1);
}

struct part_info {
  int left, length;
  part_info(int left, int length) : left(left), length(length) {}
};

template <class Combine, class Choose>
void gibbs_sampler_block::sample(sentence& s) {
  int nodes = s.nodes.size();

  static vector<part_info> parts;
  parts.clear();

  if (nodes <= MAX_LEN) parts.emplace_back(0, nodes);
  else {
    // If we have > MAX_LEN nodes, we need to find splits between sentence parts.
    // a) fill parts with overlapping parts.
    int num_parts = 1 + (nodes - 1 - OVERLAP) / (MAX_LEN - OVERLAP);
    parts.emplace_back(0, (nodes - (num_parts-2) * (MAX_LEN - OVERLAP) + OVERLAP) / 2);
    for (int part = 1; part < num_parts; part++) {
      int left = parts.back().left + parts.back().length - OVERLAP;
      parts.emplace_back(left, part < num_parts-1 ? MAX_LEN : nodes-left);
    }

    // b) compute probs of all bounds combinations between overlapping parts.
    // Rest_robs[i] is prob of all bounds combinations from word i to the end of the sentence.
    // Tree_probs[left][length] is prob of tree on length words from the i-th word.
    prob_t rest_probs[nodes+1];
    rest_probs[nodes] = 1;
    prob_t tree_probs[nodes][MAX_LEN+1];
    for (int part = num_parts-1; part >= 0; part--) {
      count_probs<Combine>(s, parts[part].left, parts[part].length);
      for (int loffset = 0; loffset < 1 + (part>0)*OVERLAP; loffset++) {
        int left = parts[part].left + loffset;
        for (int roffset = 0; roffset < 1 + (part<num_parts-1)*OVERLAP; roffset++) {
          int length = parts[part].length - roffset - loffset;
          tree_probs[left][length] = tree_prob<Combine>(loffset, parts[part].length - 1 - roffset);
          prob_t prob = tree_probs[left][length] * rest_probs[left + length];
          rest_probs[left] = roffset == 0 ? prob : Combine::apply(rest_probs[left], prob);
        }
      }
    }

    // c) choose nonoverlapping parts according to probs.
    int lbound = 0;
    for (int part = 1; part < num_parts; part++) {
      static vector<prob_t> choose_probs;
      choose_probs.clear();
      for (int rbound = parts[part].left; rbound < parts[part].left + OVERLAP; rbound++)
        choose_probs.push_back(tree_probs[lbound][rbound-lbound] * rest_probs[rbound]);

      int rbound = parts[part].left + Choose::apply(choose_probs);
      parts[part-1].length = rbound - lbound;
      parts[part].left = rbound;
      lbound = rbound;
    }
    parts[num_parts-1].length = nodes - lbound;
  }

  assert(!parts.empty());
  assert(parts.front().left == 0);
  assert(parts.back().left + parts.back().length == nodes);
  for (unsigned part = 1; part < parts.size(); part++) assert(parts[part-1].left + parts[part-1].length == parts[part].left);

  // Count probs for each sentence part and choose a tree
  s.flatten();
  foreach (auto& part, parts) {
    count_probs<Combine>(s, part.left, part.length);

    parents.resize(part.length);
    choose_root<Choose>(-1, 0, part.length-1);
    for (int n = 0; n < part.length; n++)
      s.change_parent(s.nodes[n + part.left], parents[n] >= 0 ? parents[n] + part.left : part.left == 0 ? -1 : s.root);
  }
}

template <class Combine>
void gibbs_sampler_block::count_probs(sentence& s, int from, int nodes) {
  // Precompute model probabilities
  for (int g = -1; g < nodes; g++)
    for (int d = 0; d < nodes; d++)
      if (d != g) {
        edge_probs[g][d] = 1;
        foreach (auto& m, edge_models) edge_probs[g][d] *= m->edge_prob(s, g < 0 ? -1 : g + from, d + from);
      }

  for (int r = 0; r < nodes; r++)
    for (int l = 0; l <= r; l++) {
      range_probs[l][r] = 1;
      foreach (auto& m, range_models) range_probs[l][r] *= m->range_prob(s, l + from, r + from);
    }

  for (int stop = 0; stop < 2; stop++)
    for (int g = 0; g < nodes; g++) {
      for (int dir = 0; dir < 2; dir++) {
        stop_adj_probs[stop][g][dir] = 1;
        foreach (auto& m, stop_models) stop_adj_probs[stop][g][dir] *= m->stop_prob(s, stop, g + from, dir, -1);
      }
      for (int f = 0; f < nodes; f++)
        if (f != g) {
          stop_fri_probs[stop][g][f] = 1;
          foreach (auto& m, stop_models) stop_fri_probs[stop][g][f] *= m->stop_prob(s, stop, g + from, f > g, f + from);
        }
    }

  // Init p[g][l][r] and q[g][l][r] with l == r
  for (int r = 0; r < nodes; r++)
    for (int g = 0; g < nodes; g++)
      if (r != g) {
        prob_t prob = range_probs[r][r];
        prob *= edge_probs[g][r];
        prob *= stop_adj_probs[1][r][LEFT] * stop_adj_probs[1][r][RGHT];
        p[g][r][r][0] = prob * stop_fri_probs[0][g][r];
        p[g][r][r][1] = prob * stop_fri_probs[1][g][r];
        q[g][r][r] = p[g][r][r][1];
      }

  // Compute p[g][l][r] and q[g][l][r] with increasing len = abs(r-l)
  for (int len = 1; len < nodes; len++) {
    for (int l = 0, r = len; r < nodes; l++, r++) {
      for (int g = 0; g < nodes; g++) {
        if (g == l) { g = r; continue; }

        // Compute p[g][l][r][s]
        for (int root = l; root <= r; root++) {
          prob_t prob = range_probs[l][r];
          prob *= edge_probs[g][root];

          // Left subtree
          prob *= root == l ? stop_adj_probs[1][root][LEFT] : stop_adj_probs[0][root][LEFT] * q[root][l][root-1];
          // Right subtree
          prob *= root == r ? stop_adj_probs[1][root][RGHT] : stop_adj_probs[0][root][RGHT] * q[root][root+1][r];

          // Update p[g][l][r][s]
          for (int s = 0; s < 2; s++) {
            prob_t prob_s = prob * stop_fri_probs[s][g][root];
            p[g][l][r][s] = root == l ? prob_s : Combine::apply(p[g][l][r][s], prob_s);
          }
        }

        // Compute q[g][l][r]
        q[g][l][r] = p[g][l][r][1];
        for (int sibling = l; sibling < r; sibling++) {
          prob_t prob;
          if (l > g) prob = p[g][l][sibling][0] * q[g][sibling+1][r]; // Right children of g, so one left subtree, possibly more right subtrees.
          else prob = p[g][sibling+1][r][0] * q[g][l][sibling]; // The other case.

          q[g][l][r] = Combine::apply(q[g][l][r], prob);
        }
      }
    }
  }
}

template <class Combine>
gibbs_sampler_block::prob_t gibbs_sampler_block::tree_prob(int l, int r) {
  prob_t total_prob = 0;
  for (int root = l; root <= r; root++) {
    prob_t prob = range_probs[l][r];
    prob *= edge_probs[-1][root];
    prob *= root == l ? stop_adj_probs[1][root][LEFT] : stop_adj_probs[0][root][LEFT] * q[root][l][root-1]; // Left subtree
    prob *= root == r ? stop_adj_probs[1][root][RGHT] : stop_adj_probs[0][root][RGHT] * q[root][root+1][r]; // Right subtree

    assert(isnormal(prob));
    total_prob = root == l ? prob : Combine::apply(total_prob, prob);
  }
  return total_prob;
}

template <class Choose>
void gibbs_sampler_block::choose_root(int parent, int l, int r) {
  static vector<prob_t> probs;
  probs.clear();

  for (int root = l; root <= r; root++) {
    prob_t prob = range_probs[l][r];
    prob *= edge_probs[parent][root];
    prob *= root == l ? stop_adj_probs[1][root][LEFT] : stop_adj_probs[0][root][LEFT] * q[root][l][root-1]; // Left subtree
    prob *= root == r ? stop_adj_probs[1][root][RGHT] : stop_adj_probs[0][root][RGHT] * q[root][root+1][r]; // Right subtree

    assert(isnormal(prob));
    probs.push_back(prob);
  }
  int root = l + Choose::apply(probs);
  parents[root] = parent;

  if (l < root) choose_left_children<Choose>(root, l, root-1);
  if (root < r) choose_rght_children<Choose>(root, root+1, r);
}

template <class Choose>
void gibbs_sampler_block::choose_left_children(int parent, int l, int r) {
  static vector<prob_t> probs;
  probs.clear();

  for (int bound = l; bound <= r; bound++) {
    prob_t prob = p[parent][bound][r][bound==l];
    if (bound != l) prob *= q[parent][l][bound-1];

    assert(isnormal(prob));
    probs.push_back(prob);
  }
  int bound = l + Choose::apply(probs);

  if (l < bound) choose_left_children<Choose>(parent, l, bound-1);
  choose_root<Choose>(parent, bound, r);
}

template <class Choose>
void gibbs_sampler_block::choose_rght_children(int parent, int l, int r) {
  static vector<prob_t> probs;
  probs.clear();

  for (int bound = r; bound >= l; bound--) {
    prob_t prob = p[parent][l][bound][bound==r];
    if (bound != r) prob *= q[parent][bound+1][r];

    assert(isnormal(prob));
    probs.push_back(prob);
  }
  int bound = r - Choose::apply(probs);

  if (bound < r) choose_rght_children<Choose>(parent, bound+1, r);
  choose_root<Choose>(parent, l, bound);
}

gibbs_sampler_block::prob_t gibbs_sampler_block::Combine_max::apply(prob_t a, prob_t b) {
  return max(a, b);
}

int gibbs_sampler_block::Choose_max::apply(const vector<prob_t>& probs) {
  int best = 0;
  for (unsigned i = 1; i < probs.size(); i++)
    if (probs[i] > probs[best])
      best = i;
  return best;
}

gibbs_sampler_block::prob_t gibbs_sampler_block::Combine_rnd::apply(prob_t a, prob_t b) {
  return a + b;
}

int gibbs_sampler_block::Choose_rnd::apply(const vector<prob_t>& probs) {
  prob_t sum = 0;
  foreach (auto& p, probs) sum += p;

  prob_t guess = distribution::real(0, (double) sum);

  unsigned chosen = 0;
  while (chosen+1 < probs.size() && guess > probs[chosen]) guess -= probs[chosen++];
  return chosen;
}
