// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include "common.h"
#include "gibbs_sampler_incremental.h"
#include "utils/distribution.h"

void gibbs_sampler_incremental::construct(const vector<model*>& models, bool verbose) {
  sampling_space.construct(models, verbose);
}

void gibbs_sampler_incremental::sample(sentence& s, int iter) {
  distribution::permutation(0, s.nodes.size(), perm);
  foreach (auto& ni, perm) {
    node& n = s.nodes[ni];
    if (n.parent < 0) continue;

    samples.clear();
    double prob = 1., prob_sum = 0.;
    sampling_space::sample_type sample = sampling_space.init(s, n);
    do samples.emplace_back(sample, (prob_sum += prob));
    while (sampling_space.next(sample, prob));

    if (!isnormal(prob))
      cerr << "\nWarn (Iter " << iter+1 << ", words " << s.nodes.size() << "): Prob overflow (" << prob << ")... ";
    else if (!(prob >= 1-1e6 && prob <= 1+1e6))
      cerr << "\nWarn (Iter " << iter+1 << ", words " << s.nodes.size() << "): Prob not 1 (" << prob << ")... ";
    else if (!isfinite(prob_sum))
      cerr << "\nWarn (Iter " << iter+1 << ", words " << s.nodes.size() << "): Prob_sum overflow (" << prob_sum << ")... ";
    else {
      assert(prob_sum > 0);
      prob = distribution::real(0, prob_sum);
      assert(prob < prob_sum);

      unsigned chosen = 0;
      while (samples[chosen].second < prob) chosen++;
      sampling_space.choose(samples[chosen].first);
    }
  }
}

////////// Sampling space //////////

void gibbs_sampler_incremental::sampling_space::construct(const vector<model*>& models, bool verbose) {
  foreach (auto& m, models) {
    auto incremental_model = m->incremental();
    if (!incremental_model) continue;

    incremental_models.push_back(incremental_model);

    auto comp_sampling_model = incremental_model->comp_sampling();
    if (comp_sampling_model) incremental_comp_sampling_models.push_back(comp_sampling_model);
  }

  if (verbose && !incremental_comp_sampling_models.empty()) cerr << "Using comp sampling." << endl;
}

gibbs_sampler_incremental::sampling_space::sample_type gibbs_sampler_incremental::sampling_space::init(sentence& s, node& n) {
  assert(n.parent >= 0);

  this->s = &s;
  n_lower = n.ord;
  n_upper = n.parent;

  int ndir = n.ord > n.parent;
  ch_lower[OUTER] = 0;
  for (int sibling = n.sibling[ndir]; sibling >= 0; sibling = s.nodes[sibling].sibling[ndir]) ch_lower[OUTER]++;
  ch_lower[INNER] = s.nodes[n.parent].children[ndir] - 1 - ch_lower[OUTER];
  ch_lower[OUTER] += n.children[ndir];
  ch_lower[INNER] += n.children[1^ndir]; // negace
  ch_upper[INNER] = ch_lower[INNER];
  ch_upper[OUTER] = s.nodes[n.parent].children[1^ndir];

  // Normalize the position and store the probability.
  while (n.children[ndir] < ch_lower[OUTER]) change_parent(n.sibling[ndir], n.ord);
  checkpoint_prob = 1.;

  initial = {ndir, n.children[ndir], n.children[1^ndir], 0};
  return initial;
}

bool gibbs_sampler_incremental::sampling_space::next(gibbs_sampler_incremental::sampling_space::sample_type& sample, double& prob) {
  int dir = sample.dir;
  int ch_outer = sample.ch_outer;
  int ch_inner = sample.ch_inner;

  if (ch_outer) {
    ch_outer--;
    change_parent(s->nodes[n_lower].outer_child[dir], n_upper, &prob);
  } else {
    for (; ch_outer < ch_lower[OUTER]; ch_outer++) change_parent(s->nodes[n_lower].sibling[dir], n_lower);
    prob = checkpoint_prob;
    if (ch_inner < ch_lower[INNER]) {
      ch_inner++;
      change_parent(s->nodes[n_lower].sibling[1^dir], n_lower, &prob);
    } else {
      if (dir == LEFT) for (; ch_inner > 0; ch_inner--) change_parent(s->nodes[n_lower].outer_child[1^dir], n_upper, &prob);
      // Swap lower and upper
      change_parent(n_lower, s->nodes[n_upper].parent, &prob);
      change_parent(n_upper, n_lower, &prob);
      ch_outer = ch_upper[OUTER];
      ch_inner = ch_upper[INNER] - ch_inner;
      dir = 1^dir;
      swap(n_lower, n_upper);
      swap(ch_lower[OUTER], ch_upper[OUTER]);
      swap(ch_lower[INNER], ch_upper[INNER]);
      if (dir == RGHT) for (; ch_inner > 0; ch_inner--) change_parent(s->nodes[n_lower].outer_child[1^dir], n_upper, &prob);
    }
    checkpoint_prob = prob;
  }

  sample.dir = dir;
  sample.ch_outer = ch_outer;
  sample.ch_inner = ch_inner;

  if (dir != initial.dir || ch_outer != initial.ch_outer || ch_inner != initial.ch_inner) return true;

  if (!incremental_comp_sampling_models.empty() && n_upper == s->root) {
    foreach (auto& csm, incremental_comp_sampling_models)
      csm->change_comp(*s, &prob);
    s->comp ^= 1;
    sample.comp ^= 1;

    assert(ch_outer == ch_lower[OUTER]);
    checkpoint_prob = prob;
  }
  return sample.comp != 0;
}

void gibbs_sampler_incremental::sampling_space::choose(sample_type target) {
  int dir = n_lower > n_upper;
  if (dir != target.dir) {
    assert(s->nodes[n_lower].children[dir] == ch_lower[OUTER]);
    change_parent(n_lower, s->nodes[n_upper].parent);
    change_parent(n_upper, n_lower);
    dir = 1^dir;
    swap(n_lower, n_upper);
  }
  while (s->nodes[n_lower].children[dir] < target.ch_outer) change_parent(s->nodes[n_lower].sibling[dir], n_lower);
  while (s->nodes[n_lower].children[dir] > target.ch_outer) change_parent(s->nodes[n_lower].outer_child[dir], n_upper);
  while (s->nodes[n_lower].children[1^dir] < target.ch_inner) change_parent(s->nodes[n_lower].sibling[1^dir], n_lower);
  while (s->nodes[n_lower].children[1^dir] > target.ch_inner) change_parent(s->nodes[n_lower].outer_child[1^dir], n_upper);
}

void gibbs_sampler_incremental::sampling_space::change_parent(int n, int new_parent, double* prob) {
  node& n_node = s->nodes[n];
  foreach (auto& m, incremental_models)
    m->change_parent(*s, n_node, new_parent, prob);

  s->change_parent(n_node, new_parent);
}
