// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include "common.h"
#include "directed_mst.h"

template<class T>
union edge_info {
  T w;
  int original;
};

template<class T>
struct vertex_info {
  vector<edge_info<T>> preds;
  int active;
  int cycle_head;
};

template<class T>
void directed_mst(const vector<vector<T>>& w, vector<int>& parent) {
  int N = w.size();

  // Initialize local structures.
  parent.clear(); parent.reserve(2*N); parent.resize(N, -1);
  vector<vertex_info<T>> vs; vs.reserve(2*N); vs.resize(N);
  for (int v = 0; v < N; v++) {
    vs[v].active = 1;
    vs[v].preds.reserve(2*N); vs[v].preds.resize(N);
    for (int g = 0; g < N; g++)
      vs[v].preds[g].w = w[v][g];
  }
  int cycle[N];
  int cycle_len;

  // Run Chu Liu - Edmonds -- contract
  for (int v = 1; v < N; v++)
    if (vs[v].active) {
      // Find minimum incoming edge
      int best_g = 0;
      for (int g = 1; g < N; g++)
        if (vs[g].active && g != v)
          if (vs[v].preds[g].w < vs[v].preds[best_g].w)
            best_g = g;
      parent[v] = best_g;

      // Cycle detection
      cycle_len = 0;
      cycle[cycle_len++] = v;
      T cycle_min = vs[v].preds[parent[v]].w;
      for (int pred = parent[v]; parent[pred] >= 0 && pred != v; pred = parent[pred]) {
        cycle[cycle_len++] = pred;
        if (vs[pred].preds[parent[pred]].w < cycle_min) cycle_min = vs[pred].preds[parent[pred]].w;
      }
      if (parent[cycle[cycle_len-1]] != v) continue;
      cycle[cycle_len] = v;

      // Contract cycle into new vertex N
      vertex_info<T>& z = (vs.emplace_back(), vs.back());
      z.active = 1;
      z.cycle_head = v;
      z.preds.resize(N+1);
      parent.push_back(-1);

      // Deactivate original vertices
      for (int i = 0; i < cycle_len; i++) vs[cycle[i]].active = 0;

      // Edges leaving cycle
      for (int w = 1; w < N; w++)
        if (vs[w].active) {
          vs[w].preds.emplace_back();
          for (int i = 0; i < cycle_len; i++) {
            if (!i || vs[w].preds[cycle[i]].w < vs[w].preds[N].w) {
              vs[w].preds[N].w = vs[w].preds[cycle[i]].w;
              vs[w].preds[v].original = cycle[i];
            }
            if (parent[w] == cycle[i]) parent[w] = N;
          }
        }

      // Edges entering the cycle
      for (int w = 0; w < N; w++)
        if (vs[w].active)
          for (int i = 0; i < cycle_len; i++)
            if (!i || vs[cycle[i]].preds[w].w + cycle_min - vs[cycle[i]].preds[cycle[i+1]].w < z.preds[w].w) {
              z.preds[w].w = vs[cycle[i]].preds[w].w + cycle_min - vs[cycle[i]].preds[cycle[i+1]].w;
              vs[v].preds[w].original = cycle[i];
            }

      N++;
    }

  // Expand cycles
  while(--N >= (int)w.size()) {
    int v = vs[N].cycle_head;

    // Activity of vertices
    vs[v].active = 1;
    for (int pred = parent[v]; pred != v; pred = parent[pred])
      vs[pred].active = 1;

    // Edges leaving the cycle
    for (int w = 1; w < N; w++)
      if (vs[w].active && parent[w] == N)
        parent[w] = vs[w].preds[v].original;

    // Edges entering the cycle
    parent[vs[v].preds[parent[N]].original] = parent[N];
  }
  parent.resize(w.size());
}

template void directed_mst(const vector<vector<int>>& w, vector<int>& parent);
template void directed_mst(const vector<vector<float>>& w, vector<int>& parent);
