// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include <cmath>
#include <cwctype>
#include <fstream>

#include "common.h"
#include "attach.h"
#include "utils/utils.h"

int attach_model::counts_info::tags;

attach_model::attach_model(const model_options& options) : alpha(50), use_cross(false), use_comp(false), use_fg(false), use_tg(true), unigram_base(false), use_p0(true), beta(0) {
  foreach (auto& option, options) {
    if (option.first == "" || option.first == "a" || option.first == "alpha") alpha = stod(option.second);
    else if (option.first == "b" || option.first == "beta") beta = stod(option.second);
    else if (option.first == "cross") use_cross = stoi(option.second) > 0;
    else if (option.first == "comp") use_comp = stoi(option.second) > 0;
    else if (option.first == "tg") use_tg = stoi(option.second) > 0;
    else if (option.first == "fg") use_fg = stoi(option.second) > 0;
    else if (option.first == "unigram") unigram_base = stoi(option.second) > 0;
    else if (option.first == "p0") use_p0 = stoi(option.second) > 0;
    else if (option.first == "formprobs") formprobs_file = option.second;
    else runtime_errorf("Unknown attach_model option %s", option.first.c_str());
  }
}

void attach_model::construct(const corpus& c) {
  tags = c.tags.size();
  forms = c.lcforms.size();
  counts_info::tags = tags;

  int words = 0;
  tag_prob.resize(tags);
  foreach (auto& s, c.sentences)
    foreach (auto& n, s.nodes) {
      tag_prob[n.tag]++;
      words++;
    }
  foreach (auto& p, tag_prob)
    p = unigram_base ? (p / (double) words) : (1 / (double) tags);

  if (use_cross) {
    nearest_punctuation.resize(c.sentences.size());
    foreach (auto& s, c.sentences)
      foreach (auto& n, s.nodes) {
        bool is_punct = false;
        if (!n.ord)
          nearest_punctuation[s.ord].push_back(-1);
        else {
          vector<wchar_t> wchars;
          to_wchars(c.forms[s.nodes[n.ord - 1].form], wchars);

          is_punct = true;
          for (unsigned i = 0; is_punct && i < wchars.size() - 1/*null character at the end*/; i++)
            is_punct = is_punct && iswpunct(wchars[i]);

          nearest_punctuation[s.ord].push_back(is_punct ? n.ord - 1 : nearest_punctuation[s.ord].back());
        }
      }
  }

#define formprobs_index(gf, df) (forms * (gf) + (df))

  if (beta) {
    formprobs.resize(forms * forms);
    if (!formprobs_file.empty()) {
      ifstream f(formprobs_file.c_str());
      if (!f.is_open()) runtime_errorf("Cannot open file '%s' with the form probabilities.", formprobs_file.c_str());

      str line;
      vector<str> tokens;
      while (gettextline(f, line)) {
        split(line, '\t', tokens);
        if (tokens.size() < 3) runtime_errorf("The line '%s' in form probabilities file does not contain at least three columns", line.c_str());
        auto g_form_it = c.lcforms_map.find(tokens[0]);
        if (g_form_it == c.lcforms_map.end()) continue;
        auto d_form_it = c.lcforms_map.find(tokens[1]);
        if (d_form_it == c.lcforms_map.end()) continue;
        formprobs[formprobs_index(g_form_it->second, d_form_it->second)] = stod(tokens[2]);
      }
    }
  }
}

#define counts_index(dir, tg, fg, cross) ((dir) + counts_index2(2, tg, fg, cross))
#define counts_index2(mul, tg, fg, cross) (use_tg ? (mul*(tg) + counts_index3(mul * (tags+1), fg, cross)) : (counts_index3(mul, fg, cross)))
#define counts_index3(mul, fg, cross) (use_fg ? (mul*(fg) + counts_index4(mul * (forms+1), cross)) : (counts_index4(mul, cross)))
#define counts_index4(mul, cross) (use_comp ? mul*(cross) : 0)

void attach_model::init_counts() {
  counts.clear();
  big_counts.clear();

  if (!use_fg) counts.resize(1 + counts_index(1, tags, forms, 1));
}

void attach_model::count_edge(int delta, const sentence& s, const node &n, const node *p, double *prob, int swap_comp) {
  int dir = p ? n.ord > p->ord : (use_comp ? s.comp^swap_comp : 0);
  int cross = use_cross && p ? nearest_punctuation[s.ord][max(n.ord, p->ord)] > min(n.ord, p->ord) : 0;
  int tg = p ? 1 + p->tag : 0;
  int fg = p ? 1 + p->lcform : 0;
  int index = counts_index(dir, tg, fg, cross);
  counts_info& c = use_fg ? big_counts[index] : counts[index];

  if (delta < 0) {
    c.count += delta;
    c.children[n.tag] += delta;
  }

  if (prob) {
    if (beta)
      multiplicative_increment(*prob, (c.children[n.tag] + alpha * tag_prob[n.tag]) / (c.count + alpha) + beta * (p ? formprobs[formprobs_index(p->lcform, n.lcform)] : 0), delta);
    else if (use_p0)
      multiplicative_increment(*prob, (c.children[n.tag] + alpha * tag_prob[n.tag]) / (c.count + alpha), delta);
    else
      multiplicative_increment(*prob, (c.children[n.tag] + alpha) / (c.count + alpha * tags), delta);
  }

  if (delta > 0) {
    c.count += delta;
    c.children[n.tag] += delta;
  }
}

void attach_model::count_sentence(const sentence &s, int delta, double *logprob) {
  double prob = 1;

  foreach (auto& n, s.nodes) {
    count_edge(delta, s, n, n.parent >= 0 ? &s.nodes[n.parent] : NULL, logprob ? &prob : NULL);
    if (logprob && prob < logprob_min_bound) *logprob += log(prob), prob = 1.;
  }

  if (logprob) *logprob += log(prob);
}

void attach_model::initialize(const corpus& c, double* logprob) {
  init_counts();

  foreach (auto& s, c.sentences)
    count_sentence(s, +1, logprob);
}

void attach_model::change_parent(const sentence& s, const node& n, int new_parent, double* prob) {
  count_edge(-1, s, n, n.parent >= 0 ? &s.nodes[n.parent] : NULL, prob);
  count_edge(+1, s, n, new_parent >= 0 ? &s.nodes[new_parent] : NULL, prob);
}

////////// Comp sampling //////////

incremental_comp_sampling_model* attach_model::comp_sampling() {
  return use_comp ? this : NULL;
}

void attach_model::change_comp(const sentence& s, double* prob) {
  assert(use_comp);
  assert(s.root >= 0 && s.nodes[s.root].parent < 0 && s.nodes[s.root].sibling[RGHT] < 0);

  count_edge(-1, s, s.nodes[s.root], NULL, prob, 0);
  count_edge(+1, s, s.nodes[s.root], NULL, prob, 1);
}

////////// Block model //////////

double attach_model::edge_prob(const sentence &s, int parent, int node) {
  double prob = 1;
  count_edge(0, s, s.nodes[node], parent >= 0 ? &s.nodes[parent] : NULL, &prob);
  return prob;
}

void attach_model::count_sentence(const sentence &s, int delta) {
  count_sentence(s, delta, NULL);
}

////////// Sampling model state //////////
void attach_model::sample_state() {
  if (sampled_counts.empty()) sampled_counts.resize(counts.size());
  else assert(counts.size() == sampled_counts.size());

  for (unsigned i = 0; i < counts.size(); i++)
    sampled_counts[i] += counts[i];
  foreach (auto& c, big_counts) {
    auto s = sampled_big_counts.find(c.first);
    if (s == sampled_big_counts.end()) sampled_big_counts.insert(c);
    else s->second += c.second;
  }
}

void attach_model::use_sampled_state() {
  assert(counts.size() == sampled_counts.size());

  for (unsigned i = 0; i < counts.size(); i++)
    counts[i] = sampled_counts[i];
  big_counts = sampled_big_counts;
}
