// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include <cmath>

#include "common.h"
#include "baseline.h"

baseline_model::baseline_model(const model_options& options) : mode(UNDEFINED) {
  foreach (auto& option, options) {
    if (option.first == "" && option.second == "random") mode = RANDOM;
    else if (option.first == "" && option.second == "left") mode = LEFT;
    else if (option.first == "" && option.second == "right") mode = RIGHT;
    else runtime_errorf("Unknown baseline_model option %s", option.first.c_str());
  }

  if (mode == UNDEFINED)
    runtime_errorf("Mode of baseline_model not specified");
}

void baseline_model::construct(const corpus& /*c*/) {
}

void baseline_model::count_edge(int delta, const node& n, const node* p, double& prob) {
  if (!p) return;

  switch(mode) {
    case RANDOM:
      multiplicative_increment(prob, 1., delta);
      break;
    case LEFT:
      if (n.ord < p->ord) multiplicative_increment(prob, 1. / (p->ord - n.ord), delta);
      else multiplicative_increment(prob, 0.1 / (n.ord - p->ord), delta);
      break;
    case RIGHT:
      if (n.ord > p->ord) multiplicative_increment(prob, 1. / (n.ord - p->ord), delta);
      else multiplicative_increment(prob, 0.1 / (p->ord - n.ord), delta);
      break;

    default:
      break;
  }
}

void baseline_model::initialize(const corpus& c, double* logprob) {
  if (!logprob) return;

  foreach (auto& s, c.sentences) {
    double prob = 1;
    foreach (auto& n, s.nodes) {
      count_edge(+1, n, n.parent >= 0 ? &s.nodes[n.parent] : NULL, prob);
      if (prob < logprob_min_bound) *logprob += log(prob), prob = 1.;
    }

    *logprob += log(prob);
  }
}

void baseline_model::change_parent(const sentence& s, const node& n, int new_parent, double* prob) {
  if (!prob) return;

  count_edge(-1, n, n.parent >= 0 ? &s.nodes[n.parent] : NULL, *prob);
  count_edge(+1, n, new_parent >= 0 ? &s.nodes[new_parent] : NULL, *prob);
}

double baseline_model::edge_prob(const sentence &s, int parent, int node) {
  double prob = 1;
  count_edge(0, s.nodes[node], parent >= 0 ? &s.nodes[parent] : NULL, prob);
  return prob;
}
