// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include <cmath>
#include <fstream>

#include "common.h"
#include "stop.h"
#include "utils/utils.h"

stop_model::stop_model(const model_options& options) : alpha(1), mult1(0.5), mult0(0), base1(0.67), base0(0.67), fromfile1(true), fromfile0(false), use_tf(true), use_tg(true), use_comp(0) {
  foreach (auto& option, options) {
    if (option.first == "" || option.first == "a" || option.first == "alpha") alpha = stod(option.second);
    else if (option.first == "m0" || option.first == "mult0") mult0 = stod(option.second);
    else if (option.first == "m1" || option.first == "mult1") mult1 = stod(option.second);
    else if (option.first == "b0" || option.first == "base0") base0 = stod(option.second);
    else if (option.first == "b1" || option.first == "base1") base1 = stod(option.second);
    else if (option.first == "f0" || option.first == "fromfile0") fromfile0 = stoi(option.second) > 0;
    else if (option.first == "f1" || option.first == "fromfile1") fromfile1 = stoi(option.second) > 0;
    else if (option.first == "tf") use_tf = stoi(option.second) > 0;
    else if (option.first == "tg") use_tg = stoi(option.second) > 0;
    else if (option.first == "comp") use_comp = stoi(option.second) > 0;
    else if (option.first == "priors") priors_file = option.second;
    else runtime_errorf("Unknown stop_model option %s", option.first.c_str());
  }
}

#define priors_index(dir, tag_f) ((dir) + 2*(tag_f))

void stop_model::construct(const corpus& c) {
  tags = c.tags.size();

  priors.resize(1 + priors_index(1, tags-1));
  if (!priors_file.empty()) {
    ifstream f(priors_file.c_str());
    if (!f.is_open()) runtime_errorf("Cannot open file '%s' with the stop prior probabilities.", priors_file.c_str());

    str line;
    vector<str> tokens;
    while (gettextline(f, line)) {
      split(line, '\t', tokens);
      if (tokens.size() < 3) runtime_errorf("The line '%s' in stop prior probabilities file does not contain at least three columns", line.c_str());

      int dir;
      if (tokens[0] == "L") dir = LEFT;
      else if (tokens[0] == "R") dir = RGHT;
      else runtime_errorf("Direction '%s' in the stop prior probabilities file not recognized", tokens[0].c_str());

      auto tag_it = c.tags_map.find(tokens[1]);
      if (tag_it == c.tags_map.end()) continue;

      int priors_index = priors_index(dir, tag_it->second);
      priors[priors_index].prob1 = stod(tokens[2]);
      priors[priors_index].specified = true;
    }
    f.close();
  }
}

#define count_index(dir, adj, tag_f, tag_g, comp) ((dir) + 2*(adj) + count_index2(4, tag_f, tag_g, comp))
#define count_index2(mul, tag_f, tag_g, comp) (use_tf ? (mul*(tag_f) + count_index3(mul * tags, tag_g, comp)) : count_index3(mul, tag_g, comp))
#define count_index3(mul, tag_g, comp) (use_tg ? (mul*(tag_g) + count_index4(mul * tags, comp)) : count_index4(mul, comp))
#define count_index4(mul, comp) (use_comp ? mul*(comp) : 0)

void stop_model::init_counts() {
  counts.clear();
  counts.resize(1 + count_index(1, 1, (tags-1), (tags-1), 1));
}

void stop_model::count_edge(int delta, const sentence& s, bool stop, int parent, int dir, int fringe, double* prob, int swap_comp) {
  assert(parent >= 0);

  if (fringe >= 0 && (fringe > parent) != dir) fringe = -1;
  int adj = fringe < 0;
  int tag_f = s.nodes[fringe >= 0 ? fringe : parent].tag;
  int index = count_index(dir, adj, tag_f, s.nodes[parent].tag, s.comp^swap_comp);

  if (delta < 0) {
    counts[index].count += delta;
    counts[index].stop[stop] += delta;
  }

  if (prob) {
    double prior1 = stop ? base1 : 1 - base1;
    double prior0 = stop ? base0 : 1 - base0;
    if (!priors_file.empty()) {
      int priors_index = priors_index(dir, s.nodes[parent].tag);
      if (!priors[priors_index].specified) priors_index = priors_index(dir, s.nodes[parent].oritag);
      if (fromfile1) prior1 = stop ? priors[priors_index].prob1 : 1 - priors[priors_index].prob1;
      if (fromfile0) prior0 = stop ? priors[priors_index].prob0 : 1 - priors[priors_index].prob0;
    }
    //if (!mult_alpha)
    //  multiplicative_increment(*prob, (counts[index].stop[stop] + prior * alpha) / (counts[index].count + alpha), delta);
    if (adj)
      multiplicative_increment(*prob, ((counts[index].stop[stop] + (stop ? base1 : 1 - base1)) / (counts[index].count + 1.) + mult1 * prior1) / (1. + mult1), delta);
    else
      multiplicative_increment(*prob, ((counts[index].stop[stop] + (stop ? base0 : 1 - base0)) / (counts[index].count + 1.) + mult0 * prior0) / (1. + mult0), delta);
  }

  if (delta > 0) {
    counts[index].count += delta;
    counts[index].stop[stop] += delta;
  }
}

void stop_model::count_sentence(const sentence &s, int delta, double *logprob) {
  double prob = 1;
  foreach (auto& n, s.nodes) {
    count_edge(delta, s, true, n.ord, LEFT, n.outer_child[LEFT], logprob ? &prob : NULL);
    count_edge(delta, s, true, n.ord, RGHT, n.outer_child[RGHT], logprob ? &prob : NULL);
    if (n.parent >= 0) {
      int dir = n.ord > n.parent;
      count_edge(delta, s, false, n.parent, dir, n.sibling[1^dir], logprob ? &prob : NULL);
    }
    if (logprob && prob < logprob_min_bound) *logprob += log(prob), prob = 1.;
  }

  if (logprob) *logprob += log(prob);
}

void stop_model::initialize(const corpus& c, double* logprob) {
  init_counts();

  foreach (auto& s, c.sentences)
    count_sentence(s, +1, logprob);
}

void stop_model::change_parent(const sentence& s, const node& n, int new_parent, double* prob) {
  if (n.parent >= 0) {
    int dir = n.ord > n.parent;

    count_edge(-1, s, false, n.parent, dir, n.sibling[1^dir], prob);
    count_edge(-1, s, n.sibling[dir] < 0, n.parent, dir, n.ord, prob);
    count_edge(+1, s, n.sibling[dir] < 0, n.parent, dir, n.sibling[1^dir], prob);
  }

  if (new_parent >= 0) {
    int dir = n.ord > new_parent;

    int inner_sibling = s.nodes[new_parent].outer_child[dir], outer_sibling = -1;
    if (inner_sibling >= 0 && (inner_sibling > n.ord) == dir) {
      assert(n.parent >= 0 && new_parent == s.nodes[n.parent].parent);

      inner_sibling = n.parent;
      outer_sibling = s.nodes[n.parent].sibling[n.ord > n.parent];
      if ((inner_sibling > n.ord) == dir) swap(inner_sibling, outer_sibling);
    }
    assert(inner_sibling < 0 || (inner_sibling > n.ord) != dir);
    assert(outer_sibling < 0 || (outer_sibling > n.ord) == dir);

    count_edge(+1, s, false, new_parent, dir, inner_sibling, prob);
    count_edge(-1, s, outer_sibling < 0, new_parent, dir, inner_sibling, prob);
    count_edge(+1, s, outer_sibling < 0, new_parent, dir, n.ord, prob);
  }
}

////////// Comp sampling //////////

incremental_comp_sampling_model* stop_model::comp_sampling() {
  return use_comp ? this : NULL;
}

void stop_model::change_comp(const sentence& s, double* prob) {
  assert(use_comp);

  foreach (const node& n, s.nodes) {
    count_edge(-1, s, true, n.ord, LEFT, n.outer_child[LEFT], prob, 0);
    count_edge(+1, s, true, n.ord, LEFT, n.outer_child[LEFT], prob, 1);
    count_edge(-1, s, true, n.ord, RGHT, n.outer_child[RGHT], prob, 0);
    count_edge(+1, s, true, n.ord, RGHT, n.outer_child[RGHT], prob, 1);
    if (n.parent >= 0) {
      int dir = n.ord > n.parent;
      count_edge(-1, s, false, n.parent, dir, n.sibling[1^dir], prob, 0);
      count_edge(+1, s, false, n.parent, dir, n.sibling[1^dir], prob, 1);
    }
  }
}

////////// Block model //////////

double stop_model::stop_prob(const sentence& s, bool stop, int parent, int dir, int fringe) {
  assert(parent >= 0);

  double prob = 1;
  count_edge(0, s, stop, parent, dir, fringe, &prob);
  return prob;
}

void stop_model::count_sentence(const sentence &s, int delta) {
  count_sentence(s, delta, NULL);
}

////////// Sampling model state //////////
void stop_model::sample_state() {
  if (sampled_counts.empty()) sampled_counts.resize(counts.size());
  else assert(counts.size() == sampled_counts.size());

  for (unsigned i = 0; i < counts.size(); i++)
    sampled_counts[i] += counts[i];
}

void stop_model::use_sampled_state() {
  assert(counts.size() == sampled_counts.size());

  for (unsigned i = 0; i < counts.size(); i++)
    counts[i] = sampled_counts[i];
}
