// This file is part of udpc.
//
// Copyright 2013 by Milan Straka <fox@ucw.cz>.
// Copyright 2013 by David Marecek <marecek@ufal.mff.cuni.cz>.
//
// Udpc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Udpc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with udpc.  If not, see <http://www.gnu.org/licenses/>.

#include <clocale>
#include <fstream>
#include <unordered_map>

#include "common.h"
#include "corpus/corpus.h"
#include "gibbs_sampler/gibbs_sampler.h"
#include "models/attach.h"
#include "models/baseline.h"
#include "models/stop.h"
#include "utils/options.h"
#include "utils/utils.h"

unordered_map<str, str> options = {
  // CONLL format
  {"conll-form-column", "1"},
  {"conll-tag-column", "4"},
  {"tagmap", ""},
  // Sampling
  {"i", "50"},  // iterations
  {"s", ""},    // samples
  {"d", "mst"}, // decoding
  {"v", ""},    // verbose
  // Models
  {"attach", ""},
  {"baseline", ""},
  {"stop", ""},
};

int main(int argc, char* argv[]) {
  setlocale(LC_CTYPE, "cs_CZ.UTF-8");

  int argi = parse_options(argc, argv, options);

  corpus c;

  // Load tagmap if specified.
  if (options.count("tagmap")) {
    ifstream file(options["tagmap"].c_str());
    if (!file.is_open()) runtime_errorf("Cannot open file '%s' with the tagmap", options["tagmap"].c_str());
    c.load_tagmap(file);
  }

  // Load from stdin unless there are parameters left on the command line.
  if (argi < argc)
    for (int i = argi; i < argc; i++) {
      ifstream file(argv[i]);
      if (!file.is_open()) runtime_errorf("Cannot open file '%s' with the corpus", argv[i]);
      c.load(file, stoi(options["conll-form-column"]), stoi(options["conll-tag-column"]));
    }
  else c.load(cin, stoi(options["conll-form-column"]), stoi(options["conll-tag-column"]));

  gibbs_sampler s(c, options.count("v"));

  if (options.count("attach")) s.add_model(new attach_model(parse_model_options(options["attach"])));
  if (options.count("baseline")) s.add_model(new baseline_model(parse_model_options(options["baseline"])));
  if (options.count("stop")) s.add_model(new stop_model(parse_model_options(options["stop"])));

  s.initialize();

  // Sampling
  vector<str> samplings;
  split(options["i"], ',', samplings);
  foreach(auto& sampling, samplings) {
    str iters = sampling;
    int samples = &sampling != &samplings.back() ? 0 : options.count("s") ? stoi(options["s"]) : max(1, stoi(iters) / 2);

    if (iters.find('/') != str::npos) {
      samples = stoi(iters.substr(iters.find('/')+1));
      iters.erase(iters.find('/'));
    }

    if (iters.find("bm") != str::npos) s.sample(gibbs_sampler::BLOCK_MAX, stoi(iters), samples);
    else if (iters.find("Bm") != str::npos) s.sample(gibbs_sampler::BLOCK_MAX_NOUPDATE, stoi(iters), samples);
    else if (iters.find("br") != str::npos) s.sample(gibbs_sampler::BLOCK_RND, stoi(iters), samples);
    else if (iters.find("Br") != str::npos) s.sample(gibbs_sampler::BLOCK_RND_NOUPDATE, stoi(iters), samples);
    else if (iters.find("i") != str::npos) s.sample(gibbs_sampler::INCREMENTAL, stoi(iters), samples);
    else runtime_errorf("Unrecognized iteration mode '%s'", iters.c_str());
  }

  // Decoding
  if (options["d"] == "mst") s.decode(gibbs_sampler::MST);
  else if (options["d"] == "counts") s.decode(gibbs_sampler::BEST_PROB);
  else runtime_errorf("Unrecognized decode mode '%s'", options["d"].c_str());

  // Save the corpus
  c.save(cout);

  return 0;
}
